@@ -69,7 +69,9 @@
 #define SCTRES_TLE	2
 #define SCTRES_OLE	3
 #define SCTRES_SLE	4	/* 暫停次數太多 */
-#define SCTRES_MAX	5
+#define SCTRES_AB	5	/* 使用者中斷 */
+#define SCTRES_UD	6	/* 未定義的東西，這應該不會出現吧 */
+#define SCTRES_MAX	7
 
 /* XXX 這種沒用的 macro 就丟了吧 */
 #define getdvalue_quotient(num, div)	((num)/(div))
@@ -103,17 +105,39 @@ static const char* childmsg_text[SCTCHILD_MSGMAX] = {
 
 /* UD (undefined) 並不會寫在這裡 */
 const char* sctres_text[2][SCTRES_MAX] = {
-	{"OK", "RE", "TLE", "OLE", "SLE"},
+	{"OK", "RE", "TLE", "OLE", "SLE", "AB", "UD"},
 	{
 		"正常結束",
 		"執行過程中發生錯誤",
 		"超過時間限制",
 		"輸出超過限制",
-		"暫停次數太多"
+		"暫停次數太多",
+		"使用者中斷操作",
+		"未定義的結果"
 	}
 };
 
-static void sctjudge_makechild_cleanup(mcopt, oldperm, copiedexe)
+static void sctjudge_makechild_cleanup_p1(void){
+	pthread_mutex_lock(&tkill_mx);
+	if(tkill_yes){
+		tkill_yes = 0;
+		pthread_mutex_unlock(&tkill_mx);
+		pthread_cancel(tkill);
+	}else{
+		pthread_mutex_unlock(&tkill_mx);
+	}
+
+	pthread_mutex_lock(&tdisplay_mx);
+	if(tdisplay_yes){
+		tdisplay_yes = 0;
+		pthread_mutex_unlock(&tdisplay_mx);
+		pthread_cancel(tdisplay);
+	}else{
+		pthread_mutex_unlock(&tdisplay_mx);
+	}
+}
+	
+static void sctjudge_makechild_cleanup_p2(mcopt, oldperm, copiedexe)
 	const struct makechildopt* mcopt;
 	const mode_t oldperm;
 	char* copiedexe;
@@ -132,24 +156,15 @@ static void sctjudge_makechild_cleanup(mcopt, oldperm, copiedexe)
 		unlink(copiedexe);
 		free(copiedexe);
 	}
+}
 
-	pthread_mutex_lock(&tkill_mx);
-	if(tkill_yes){
-		tkill_yes = 0;
-		pthread_mutex_unlock(&tkill_mx);
-		pthread_cancel(tkill);
-	}else{
-		pthread_mutex_unlock(&tkill_mx);
-	}
-
-	pthread_mutex_lock(&tdisplay_mx);
-	if(tdisplay_yes){
-		tdisplay_yes = 0;
-		pthread_mutex_unlock(&tdisplay_mx);
-		pthread_cancel(tdisplay);
-	}else{
-		pthread_mutex_unlock(&tdisplay_mx);
-	}
+static void sctjudge_makechild_cleanup(mcopt, oldperm, copiedexe)
+	const struct makechildopt* mcopt;
+	const mode_t oldperm;
+	char* copiedexe;
+{
+	sctjudge_makechild_cleanup_p1();
+	sctjudge_makechild_cleanup_p2(mcopt, oldperm, copiedexe);
 }
 
 static int copyfilebyname(const char* src, const char* destdir
@@ -370,16 +385,24 @@ void* sctjudge_makechild(void* arg){
 	struct rusage childrusage;
 	int childexitstat;
 	char childterminated = 0;
+#ifndef HAVE_LINUX_EXTENSIONS
+	enable_setuid();
+#endif
 	pipe(childpipe);
 	pthread_mutex_lock(&pidmutex);
 	pidchild = fork();
 	if(pidchild < 0){
+#ifndef HAVE_LINUX_EXTENSIONS
+		disable_setuid();
+#endif
 		pthread_mutex_unlock(&pidmutex);
 		fprintf(stderr, "子程序建立失敗：%s\n", strerror(errno));
 		abort_makechild(3);
 	}
 	else if(pidchild > 0){
 		/* 我們原本的 thread */
+		/* 從這裡開始，我們需要 setuid 來 kill 子程序，
+		 * 所以說不要執行 disable_setuid() */
 		pidcopy = pidchild;
 		pthread_mutex_unlock(&pidmutex);
 		close(childpipe[1]);
@@ -463,19 +486,22 @@ void* sctjudge_makechild(void* arg){
 		}
 		pthread_mutex_unlock(&judge_tle_mx);
 		difftimespec(&execstart, &execfinish, &judge_time);
+		if(break_flag){
+			judge_result = SCTRES_AB;
+		}
 		/* 這裡就要顯示結果了 
 		 * XXX 應該拿去 main 放的，用個 struct 包起來？ */
+		sctjudge_makechild_cleanup_p1();
 		if(mcopt.flags & SCTMC_VERBOSE){
 			putchar('\r');
-			puts("========================================");
+			puts("=======================================================");
 			fputs("評測結果：", stderr);
-			if(judge_result < 0 || judge_result > SCTRES_MAX){
-				puts("UD (未定義的結果)");
-			}else{
-				printf("%s (%s)\n", 
-						sctres_text[0][judge_result],
-						sctres_text[1][judge_result]);
+			if(judge_result < 0 || judge_result >= SCTRES_UD){
+				judge_result = SCTRES_UD;
 			}
+			printf("%s (%s)\n", 
+					sctres_text[0][judge_result],
+					sctres_text[1][judge_result]);
 			printf("執行時間：%ld.%03ld 秒\n",
 					judge_time.tv_sec,
 					getdvalue_quotient(judge_time.tv_nsec, 1000000));
@@ -483,17 +509,16 @@ void* sctjudge_makechild(void* arg){
 			if(judge_result == SCTRES_RE){
 				printf("程式因訊號 %d 而終止\n", judge_signal);
 			}
-			puts("========================================");
+			puts("=======================================================");
 		}else{
 			/* 輸出格式：
 			 * 第一行是結果代碼、時間、離開狀態 
 			 * 第二行是收到的訊號 (只在 RE 時會有) */
-			if(judge_result < 0 || judge_result > SCTRES_MAX){
-				fputs("UD ", stdout);
-			}else{
-				fputs(sctres_text[0][judge_result], stdout);
-				putchar(' ');
+			if(judge_result < 0 || judge_result >= SCTRES_UD){
+				judge_result = SCTRES_UD;
 			}
+			fputs(sctres_text[0][judge_result], stdout);
+			putchar(' ');
 			printf("%ld%03ld %d\n", judge_time.tv_sec,
 					getdvalue_quotient(judge_time.tv_nsec, 1000000),
 					judge_exitcode);
@@ -513,6 +538,13 @@ void* sctjudge_makechild(void* arg){
 
 		/* 重設所有 signal handler
 		 * XXX 我猜 signal 最大值是 32，還有其他更好的寫法嗎？ */
+		for(i=1; i<=32; i++){
+			signal(i, SIG_DFL);
+		}
+
+#ifndef HAVE_LINUX_EXTENSIONS
+		disable_setuid();
+#endif
 
 		fdinput = open(mcopt.inputfile, O_RDONLY);
 		if(fdinput < 0){
@@ -544,6 +576,11 @@ void* sctjudge_makechild(void* arg){
 			SCTCHILD_WRITE_SUCCMSG(SCTCHILD_DUP2_OUTPUT);
 		}
 
+#ifndef HAVE_LINUX_EXTENSIONS
+		enable_setuid();
+#endif
+		fchown(fdoutput, getuid(), -1);
+
 		/* 再來就是 stderr 了 */
 		if(mcopt.flags & SCTMC_REDIR_STDERR){
 			if(dup2(fdoutput, STDERR_FILENO) < 0){
@@ -707,7 +744,7 @@ void* sctjudge_makechild(void* arg){
 		exit(2);
 	}
 	/* Judge 完成，準備離開 */
-	sctjudge_makechild_cleanup(&mcopt, chrdir_oldmode, execdestname);
+	sctjudge_makechild_cleanup_p2(&mcopt, chrdir_oldmode, execdestname);
 	(*(int*)toreturn) = 0;
 	return toreturn;
 }
