@@ -1,6 +1,7 @@
 #ifndef SCTJUDGE_JUDEGCORE_HEADER
 #define SCTJUDGE_JUDEGCORE_HEADER
 
+#include <signal.h>
 #include <unistd.h>
 #include <pthread.h>
 #include <semaphore.h>
@@ -10,6 +11,7 @@ extern pthread_mutex_t pidmutex, tkill_mx, tdisplay_mx, judge_tle_mx;
 extern pthread_t tkill, tdisplay;
 extern char tkill_yes, tdisplay_yes, judge_tle;
 extern sem_t mcthr, addthr;
+extern volatile sig_atomic_t break_flag;
 
 #define SCTMC_REDIR_STDERR 0x00000001
 #define SCTMC_NOCOPY       0x00000002
